from facenet_pytorch import MTCNN, InceptionResnetV1
import torch
from PIL import Image
import numpy as np

class FaceDetector:
    def __init__(self):
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.mtcnn = MTCNN(keep_all=True, device=self.device)
        self.resnet = InceptionResnetV1(pretrained='vggface2').eval().to(self.device)

    def detect_faces(self, image_path: str):
        try:
            img = Image.open(image_path).convert('RGB')
            # Detect faces
            boxes, probs = self.mtcnn.detect(img)
            
            faces_data = []
            if boxes is not None:
                # Get embeddings
                faces_aligned = self.mtcnn(img)
                if faces_aligned is not None:
                    embeddings = self.resnet(faces_aligned.to(self.device)).detach().cpu().numpy()
                    
                    for i, box in enumerate(boxes):
                        faces_data.append({
                            "box": box.tolist(),
                            "prob": float(probs[i]),
                            "embedding": embeddings[i].tolist()
                        })
            return faces_data
        except Exception as e:
            print(f"Error detecting faces in {image_path}: {e}")
            return []

    def cluster_faces(self, all_faces_data, threshold=0.8):
        """
        Groups faces by similarity.
        all_faces_data: List of dicts, each containing 'embedding' (list) and 'image_path'.
        Returns: List of clusters (each cluster is a list of image_paths).
        """
        if not all_faces_data:
            return []
            
        embeddings = torch.tensor([f['embedding'] for f in all_faces_data]).to(self.device)
        n = len(embeddings)
        
        # Calculate distance matrix
        dists = torch.cdist(embeddings, embeddings)
        
        # Simple greedy clustering
        clusters = []
        visited = set()
        
        for i in range(n):
            if i in visited:
                continue
                
            # Start a new cluster with this face
            current_cluster = [all_faces_data[i]]
            visited.add(i)
            
            # Find all similar faces
            for j in range(i + 1, n):
                if j in visited:
                    continue
                
                if dists[i, j] < threshold:
                    current_cluster.append(all_faces_data[j])
                    visited.add(j)
            
            clusters.append(current_cluster)
            
        return clusters



import sqlite3
import json
import os
from typing import Dict, Any, Optional

class Database:
    def __init__(self, db_path: str = "video_highlight_generator.db"):
        self.db_path = db_path
        self._init_db()

    def _init_db(self):
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        
        # Images table
        cursor.execute('''
            CREATE TABLE IF NOT EXISTS images (
                path TEXT PRIMARY KEY,
                score REAL,
                tags TEXT,
                faces TEXT,
                timestamp REAL
            )
        ''')
        
        # Persons table (for clustering)
        cursor.execute('''
            CREATE TABLE IF NOT EXISTS persons (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                name TEXT,
                face_encoding TEXT
            )
        ''')
        
        conn.commit()
        conn.close()

    def get_image_data(self, path: str) -> Optional[Dict[str, Any]]:
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        cursor.execute("SELECT score, tags, faces FROM images WHERE path = ?", (path,))
        row = cursor.fetchone()
        conn.close()
        
        if row:
            return {
                "score": row[0],
                "tags": json.loads(row[1]),
                "faces": json.loads(row[2])
            }
        return None

    def save_image_data(self, path: str, score: float, tags: list, faces: list):
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        cursor.execute('''
            INSERT OR REPLACE INTO images (path, score, tags, faces, timestamp)
            VALUES (?, ?, ?, ?, ?)
        ''', (path, score, json.dumps(tags), json.dumps(faces), os.path.getmtime(path)))
        conn.commit()
        conn.close()

    def get_images_batch(self, paths: list) -> Dict[str, Dict[str, Any]]:
        if not paths:
            return {}
        
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        
        # SQLite has a limit on variables, so we might need to chunk if paths is huge
        # But for typical usage < 999 is fine. 
        # To be safe, let's just fetch all and filter in python or use chunks.
        # Simplest for now: Fetch all images that match.
        
        placeholders = ','.join(['?'] * len(paths))
        query = f"SELECT path, score, tags, faces FROM images WHERE path IN ({placeholders})"
        
        try:
            cursor.execute(query, paths)
            rows = cursor.fetchall()
            results = {}
            for row in rows:
                results[row[0]] = {
                    "score": row[1],
                    "tags": json.loads(row[2]),
                    "faces": json.loads(row[3])
                }
            return results
        except sqlite3.OperationalError:
            # Fallback for too many variables: chunk it
            results = {}
            chunk_size = 900
            for i in range(0, len(paths), chunk_size):
                chunk = paths[i:i + chunk_size]
                placeholders = ','.join(['?'] * len(chunk))
                query = f"SELECT path, score, tags, faces FROM images WHERE path IN ({placeholders})"
                cursor.execute(query, chunk)
                for row in cursor.fetchall():
                    results[row[0]] = {
                        "score": row[1],
                        "tags": json.loads(row[2]),
                        "faces": json.loads(row[3])
                    }
            return results
        finally:
            conn.close()

    def save_images_batch(self, data_list: list):
        # data_list is list of tuples: (path, score, tags, faces)
        if not data_list:
            return
            
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        
        formatted_data = []
        for item in data_list:
            path, score, tags, faces = item
            formatted_data.append((path, score, json.dumps(tags), json.dumps(faces), os.path.getmtime(path)))
            
        cursor.executemany('''
            INSERT OR REPLACE INTO images (path, score, tags, faces, timestamp)
            VALUES (?, ?, ?, ?, ?)
        ''', formatted_data)
        
        conn.commit()
        conn.close()



import torch
from torchvision import models, transforms
from PIL import Image
import cv2
import numpy as np
import os

class ImageProcessor:
    def __init__(self):
        # Load pre-trained MobileNetV3
        self.model = models.mobilenet_v3_small(pretrained=True)
        self.model.eval()
        
        # Standard ImageNet normalization
        self.preprocess = transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ])
        
        # Load ImageNet labels (simplified for this example)
        self.labels = self._load_labels()

    def _load_labels(self):
        # Placeholder for actual ImageNet labels
        # In a real app, you'd load these from a file
        return {i: f"class_{i}" for i in range(1000)}

    def process_image(self, image_path):
        try:
            # Handle non-ASCII paths for OpenCV
            with open(image_path, "rb") as f:
                file_bytes = np.frombuffer(f.read(), dtype=np.uint8)
                cv_img = cv2.imdecode(file_bytes, cv2.IMREAD_COLOR)
            
            if cv_img is None:
                return None

            # Calculate sharpness using Laplacian
            gray = cv2.cvtColor(cv_img, cv2.COLOR_BGR2GRAY)
            sharpness = cv2.Laplacian(gray, cv2.CV_64F).var()
            
            # Prepare for PyTorch
            # Convert BGR (OpenCV) to RGB (PIL)
            img_rgb = cv2.cvtColor(cv_img, cv2.COLOR_BGR2RGB)
            pil_img = Image.fromarray(img_rgb)
            
            input_tensor = self.preprocess(pil_img)
            input_batch = input_tensor.unsqueeze(0) # create a mini-batch as expected by the model

            with torch.no_grad():
                output = self.model(input_batch)
            
            probabilities = torch.nn.functional.softmax(output[0], dim=0)
            score = probabilities.max().item()
            
            # Get top predicted class as tag (placeholder logic)
            top_prob, top_catid = torch.topk(probabilities, 1)
            
            # Combine confidence and sharpness for a quality score
            # Normalize sharpness roughly (0-1000 to 0-1)
            final_score = score * 0.7 + (min(sharpness, 1000) / 1000) * 0.3
            
            return {
                "score": final_score,
                "tags": [str(top_catid.item())], # In real app, map to text
                "sharpness": sharpness
            }
        except Exception as e:
            print(f"Error analyzing {image_path}: {e}")
            return None



import os
import uvicorn
from fastapi import FastAPI, BackgroundTasks
from fastapi.staticfiles import StaticFiles
from fastapi.responses import FileResponse
import mimetypes

mimetypes.add_type("application/javascript", ".js")
mimetypes.add_type("text/css", ".css")
mimetypes.add_type("video/mp4", ".mp4")

app = FastAPI(title="Video Highlight Generator")

from pydantic import BaseModel
from typing import List, Optional
from video_highlight_generator.core.image_processor import ImageProcessor
from video_highlight_generator.core.video_generator import VideoGenerator, get_image_date
from video_highlight_generator.core.database import Database
from video_highlight_generator.core.face_detector import FaceDetector
import glob

# Initialize Core Modules
# Initialize Core Modules
db_path = os.getenv("DB_PATH", "video_highlight_generator.db")
db = Database(db_path)
image_processor = ImageProcessor()
video_generator = VideoGenerator()
face_detector = FaceDetector()

class AnalysisRequest(BaseModel):
    folder_paths: List[str]

class VideoRequest(BaseModel):
    image_paths: List[str]
    audio_path: Optional[str] = None
    audio_start: Optional[float] = 0.0
    audio_end: Optional[float] = None
    output_path: str
    resolution: str = "1080p" # 1080p (PC) or 9:16 (Phone)
    image_duration: float = 3.0
    title_text: Optional[str] = ""
    ken_burns_effect: bool = False

from fastapi import Header
from fastapi.responses import StreamingResponse
import aiofiles

async def chunk_generator(file_path, start, end, chunk_size=1024*1024):
    try:
        async with aiofiles.open(file_path, "rb") as f:
            await f.seek(start)
            while start <= end:
                read_size = min(chunk_size, end - start + 1)
                data = await f.read(read_size)
                if not data:
                    break
                yield data
                start += len(data)
    except (ConnectionResetError, OSError):
        # Client disconnected, stop streaming
        pass

@app.get("/api/image")
async def get_image(path: str, range: str = Header(None)):
    print(f"Requested media path: {path}")
    if not os.path.exists(path):
        print(f"Media not found: {path}")
        return {"error": "Media not found"}, 404
    
    # Guess mime type
    media_type, _ = mimetypes.guess_type(path)
    if media_type is None:
        media_type = "application/octet-stream"
        
    # For video files, use custom streaming to handle Windows asyncio errors gracefully
    if media_type.startswith("video/") and range:
        try:
            file_size = os.path.getsize(path)
            start, end = range.replace("bytes=", "").split("-")
            start = int(start)
            end = int(end) if end else file_size - 1
            
            # Ensure bounds
            if start >= file_size:
                return {"error": "Requested range not satisfiable"}, 416
            
            chunk_size = 1024 * 1024 # 1MB chunks
            
            headers = {
                "Content-Range": f"bytes {start}-{end}/{file_size}",
                "Accept-Ranges": "bytes",
                "Content-Length": str(end - start + 1),
                "Content-Type": media_type,
            }
            
            return StreamingResponse(
                chunk_generator(path, start, end, chunk_size),
                headers=headers,
                status_code=206,
            )
        except Exception as e:
            print(f"Error streaming video: {e}")
            # Fallback to FileResponse
            return FileResponse(path, media_type=media_type)
        
    return FileResponse(path, media_type=media_type)

@app.post("/api/browse")
async def browse_folder():
    try:
        import tkinter as tk
        from tkinter import filedialog
        
        root = tk.Tk()
        root.withdraw()
        root.attributes('-topmost', True)
        
        folder_path = filedialog.askdirectory()
        root.destroy()
        
        if folder_path:
            folder_path = os.path.normpath(folder_path)
            return {"path": folder_path}
        return {"path": ""}
    except Exception as e:
        print(f"Error opening dialog: {e}")
        return {"error": str(e)}, 500

@app.post("/api/browse_file")
async def browse_file():
    try:
        import tkinter as tk
        from tkinter import filedialog
        
        root = tk.Tk()
        root.withdraw()
        root.attributes('-topmost', True)
        
        file_path = filedialog.askopenfilename(filetypes=[("Audio Files", "*.mp3 *.wav *.aac *.m4a")])
        root.destroy()
        
        if file_path:
            file_path = os.path.normpath(file_path)
            return {"path": file_path}
        return {"path": ""}
    except Exception as e:
        print(f"Error opening dialog: {e}")
        return {"error": str(e)}, 500

# Global progress state
progress_state = {
    "status": "idle",
    "percent": 0,
    "message": ""
}

@app.get("/api/progress")
async def get_progress():
    return progress_state

def update_progress(status, percent, message):
    progress_state["status"] = status
    progress_state["percent"] = percent
    progress_state["message"] = message

def process_analysis(request: AnalysisRequest):
    try:
        # Validate folders
        valid_folders = []
        for path in request.folder_paths:
            if os.path.exists(path):
                valid_folders.append(path)
        
        if not valid_folders:
            update_progress("error", 0, "No valid folders found")
            return
        
        update_progress("analyzing", 0, "Starting analysis...")
        
        # Find images across all folders
        extensions = ['*.jpg', '*.jpeg', '*.png']
        image_files = []
        seen_paths = set()
        for folder in valid_folders:
            for ext in extensions:
                for img_path in glob.glob(os.path.join(folder, ext)):
                    norm_path = os.path.normpath(img_path)
                    if norm_path not in seen_paths:
                        image_files.append(norm_path)
                        seen_paths.add(norm_path)
        
        total_images = len(image_files)
        if total_images == 0:
            update_progress("completed", 100, "No images found")
            progress_state["result"] = {"results": [], "people": [], "themes": []}
            return

        # Analyze images
        results = []
        
        # Batch check DB
        update_progress("analyzing", 0, "Checking cache...")
        cached_data = db.get_images_batch(image_files)
        
        new_images_to_process = []
        
        for img_path in image_files:
            if img_path in cached_data:
                data = cached_data[img_path]
                data['path'] = img_path
                results.append(data)
            else:
                new_images_to_process.append(img_path)
                
        # Process new images
        new_results_to_save = []
        total_new = len(new_images_to_process)
        
        for i, img_path in enumerate(new_images_to_process):
            # Update progress based on new images processing
            # We could also factor in cached images but processing new ones takes the most time
            percent = int((i / total_new) * 90) if total_new > 0 else 90
            update_progress("analyzing", percent, f"Analyzing new image {i+1}/{total_new}...")
            
            data = image_processor.process_image(img_path)
            if data:
                # Detect faces
                faces = face_detector.detect_faces(img_path)
                data['faces'] = faces
                data['path'] = img_path
                
                results.append(data)
                new_results_to_save.append((img_path, data['score'], data['tags'], faces))
                
                # Save in batches of 10
                if len(new_results_to_save) >= 10:
                    db.save_images_batch(new_results_to_save)
                    new_results_to_save = []
        
        # Save remaining
        if new_results_to_save:
            db.save_images_batch(new_results_to_save)
                
        # Face Detection & Clustering
        update_progress("clustering", 90, "Grouping faces...")
        
        # Flatten all faces for clustering
        all_faces = []
        for img_data in results:
            for face in img_data.get('faces', []):
                face['image_path'] = img_data.get('path', '') # Ensure path is linked
                all_faces.append(face)
                
        people_clusters = face_detector.cluster_faces(all_faces)
        
        # Format people for frontend
        people = []
        for i, cluster in enumerate(people_clusters):
            if not cluster: continue
            # Use the first face's image as the representative
            people.append({
                "id": str(i),
                "name": f"Person {i+1}",
                "count": len(cluster),
                "images": list(set([f['image_path'] for f in cluster]))
            })
        
        # Theme Detection (Simple tag aggregation)
        themes = {}
        for img in results:
            for tag in img.get('tags', []):
                themes[tag] = themes.get(tag, 0) + 1
                
        sorted_themes = [{'name': k, 'count': v} for k, v in sorted(themes.items(), key=lambda item: item[1], reverse=True)]
        
        update_progress("completed", 100, "Analysis complete")
        progress_state["result"] = {
            "results": results,
            "people": people,
            "themes": sorted_themes
        }

    except Exception as e:
        print(f"Error in background analysis: {e}")
        update_progress("error", 0, f"Error: {str(e)}")

@app.post("/api/analyze")
async def analyze_folder(request: AnalysisRequest, background_tasks: BackgroundTasks):
    # Start background task
    background_tasks.add_task(process_analysis, request)
    return {"status": "started", "message": "Analysis started in background"}

from fastapi import BackgroundTasks

# ... (imports)

def process_video_generation(request: VideoRequest, output_path: str, resolution: tuple, final_images: list):
    try:
        def progress_callback(percentage, **kwargs):
            update_progress("generating", percentage, f"Generating video: {int(percentage)}%")

        update_progress("generating", 0, "Starting video generation...")
        
        # Fetch face data for each image
        images_with_data = []
        for path in final_images:
            data = db.get_image_data(path)
            faces = data.get('faces', []) if data else []
            images_with_data.append({'path': path, 'faces': faces})

        success = video_generator.generate_video(
            images_with_data, 
            request.audio_path, 
            output_path,
            resolution=resolution,
            audio_start=request.audio_start,
            audio_end=request.audio_end,
            image_duration=request.image_duration,
            title_text=request.title_text,
            ken_burns_effect=request.ken_burns_effect,
            progress_callback=progress_callback
        )
        
        if success:
            update_progress("completed", 100, "Video generation complete")
            progress_state["result"] = output_path
        else:
            update_progress("error", 0, "Video generation failed")
            
    except Exception as e:
        print(f"Error in background generation: {e}")
        update_progress("error", 0, f"Error: {str(e)}")

@app.post("/api/generate")
async def generate_video(request: VideoRequest, background_tasks: BackgroundTasks):
    resolution = (1920, 1080)
    if request.resolution == "9:16":
        resolution = (1080, 1920)
        
    # Smart Selection: Filter out low quality images if we have enough
    final_images = request.image_paths
    
    # Deduplicate input list just in case
    final_images = list(dict.fromkeys(final_images))

    if len(final_images) < 10:
        return {"status": "error", "message": "Not enough images selected. Please select at least 10 images."}, 400
    
    # Sort images chronologically
    try:
        final_images.sort(key=lambda x: get_image_date(x))
    except Exception as e:
        print(f"Error sorting images: {e}")
        # Fallback to name sort if date sort fails
        final_images.sort()

    if len(final_images) > 20:
        scored_images = []
        for path in final_images:
            data = db.get_image_data(path)
            score = data['score'] if data else 0
            scored_images.append({'path': path, 'score': score})
            
        # Sort by score descending to pick best ones
        scored_images.sort(key=lambda x: x['score'], reverse=True)
        
        # Keep top 80%
        cutoff = int(len(scored_images) * 0.8)
        # Get the paths of the best images
        final_images = [x['path'] for x in scored_images[:cutoff]]
        

        
    # Handle Output Path
    output_path = request.output_path
    if os.path.isdir(output_path) or not output_path.endswith(('.mp4', '.mov', '.avi')):
        # It's a directory (or user just gave a folder path), generate filename
        from datetime import datetime
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        filename = f"highlight_{timestamp}.mp4"
        if not os.path.exists(output_path):
             # If it doesn't exist and looks like a path, create it
             try:
                 os.makedirs(output_path, exist_ok=True)
             except:
                 # Fallback to current dir if invalid
                 output_path = "."
        
        output_path = os.path.join(output_path, filename)
        
    # Start background task
    background_tasks.add_task(process_video_generation, request, output_path, resolution, final_images)
    
    return {"status": "started", "message": "Video generation started in background"}

# Static Files
# Mount static directory to serve frontend assets
static_dir = os.path.join(os.path.dirname(__file__), "static")
if not os.path.exists(static_dir):
    os.makedirs(static_dir)

app.mount("/assets", StaticFiles(directory=os.path.join(static_dir, "assets")), name="assets")

# Catch-all route for SPA
@app.get("/{full_path:path}")
async def serve_spa(full_path: str):
    # API requests should not be handled here, but if they fall through, return 404
    if full_path.startswith("api"):
        return {"error": "Not Found"}, 404
    
    # Serve index.html for any other route (SPA)
    index_path = os.path.join(static_dir, "index.html")
    if os.path.exists(index_path):
        return FileResponse(index_path)
    return {"message": "Frontend not built yet. Please run 'npm run build' in frontend directory."}


def start():
    """Entry point for the console script"""
    uvicorn.run("video_highlight_generator.main:app", host="0.0.0.0", port=8000, reload=False)

if __name__ == "__main__":
    start()



from moviepy import ImageClip, concatenate_videoclips, AudioFileClip, concatenate_audioclips, vfx, CompositeVideoClip, VideoClip
# from moviepy.audio.fx.all import audio_loop # Removed in v2?
# Try to import audio_loop from where it might be, or use method
try:
    from moviepy.audio.fx.audio_loop import audio_loop
except ImportError:
    try:
        from moviepy.audio.fx.loop import loop as audio_loop
    except ImportError:
        # Fallback or maybe it's a method now
        audio_loop = None
import os
import random
from PIL import Image, ImageDraw, ImageFont
import numpy as np

import proglog

class CallbackLogger(proglog.ProgressBarLogger):
    def __init__(self, callback=None):
        super().__init__()
        self.progress_func = callback

    def bars_callback(self, bar, attr, value, old_value=None):
        if self.progress_func:
            # bar is usually 't' for time
            percentage = (value / self.bars[bar]['total']) * 100
            self.progress_func(percentage)

    def log(self, message):
        pass # Optional: log messages

from datetime import datetime
from PIL import Image, ImageDraw, ImageFont, ExifTags
import time

def get_image_date(image_path):
    try:
        image = Image.open(image_path)
        exif = image._getexif()
        if exif:
            for tag, value in exif.items():
                decoded = ExifTags.TAGS.get(tag, tag)
                if decoded == 'DateTimeOriginal' or decoded == 'DateTime':
                    return datetime.strptime(value, '%Y:%m:%d %H:%M:%S')
    except Exception as e:
        pass
        
    # Fallback to file modification time
    timestamp = os.path.getmtime(image_path)
    return datetime.fromtimestamp(timestamp)

class VideoGenerator:
    def __init__(self):
        pass

    def generate_video(self, images, audio_path, output_path, resolution=(1920, 1080), audio_start=0.0, audio_end=None, image_duration=3.0, title_text="", ken_burns_effect=False, progress_callback=None):
        try:
            clips = []
            
            # Prepare font for title if needed
            title_font = None
            date_font = None
            
            try:
                # Try to load a default font
                font_paths = [
                    "C:/Windows/Fonts/arial.ttf", # Windows
                    "C:/Windows/Fonts/seguiemj.ttf", # Windows
                    "/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf", # Linux
                    "/usr/share/fonts/truetype/liberation/LiberationSans-Regular.ttf", # Linux
                    "/System/Library/Fonts/Helvetica.ttc", # macOS
                ]
                
                font_path = None
                for path in font_paths:
                    if os.path.exists(path):
                        font_path = path
                        break
                
                if font_path:
                    print(f"Loading font from: {font_path}")
                    title_font = ImageFont.truetype(font_path, 100)
                    date_font = ImageFont.truetype(font_path, 70)
                else:
                    print("Warning: Custom fonts not found, using default (tiny) font.")
                    title_font = ImageFont.load_default()
                    date_font = ImageFont.load_default()
            except Exception as e:
                print(f"Error loading font: {e}")
                title_font = ImageFont.load_default()
                date_font = ImageFont.load_default()

            for i, img_entry in enumerate(images):
                # Handle dict or string
                if isinstance(img_entry, dict):
                    img_path = img_entry['path']
                    faces = img_entry.get('faces', [])
                else:
                    img_path = img_entry
                    faces = []

                # Create clip, resize to fit resolution (maintaining aspect ratio), and center on black background
                # Use custom duration
                duration_per_image = float(image_duration)
                
                # Get date string
                img_date = get_image_date(img_path)
                date_str = img_date.strftime("%B %d, %Y")
                
                # Open image
                pil_img = Image.open(img_path).convert("RGBA")
                w, h = pil_img.size
                target_w, target_h = resolution
                
                if ken_burns_effect:
                    # KEN BURNS LOGIC
                    # Calculate Focus Point (0.0 - 1.0)
                    focus_x, focus_y = 0.5, 0.3 # Default: Top-Center
                    
                    if faces:
                        # Calculate centroid of faces
                        sum_x = 0
                        sum_y = 0
                        count = 0
                        for face in faces:
                            box = face.get('box')
                            if box:
                                # box is [x1, y1, x2, y2]
                                cx = (box[0] + box[2]) / 2
                                cy = (box[1] + box[3]) / 2
                                sum_x += cx
                                sum_y += cy
                                count += 1
                        
                        if count > 0:
                            focus_x = (sum_x / count) / w
                            focus_y = (sum_y / count) / h
                            # Clamp
                            focus_x = max(0.0, min(1.0, focus_x))
                            focus_y = max(0.0, min(1.0, focus_y))

                    # Resize logic for Ken Burns (Cover)
                    scale = max(target_w / w, target_h / h)
                    new_w = int(w * scale)
                    new_h = int(h * scale)
                    
                    pil_img_resized = pil_img.resize((new_w, new_h), Image.Resampling.LANCZOS)
                    
                    # Crop to aspect ratio first (centered on focus point?)
                    # If we just resize to cover, we might lose the focus point if we crop blindly.
                    # But _apply_ken_burns handles the movement.
                    # We should pass the 'cover' resized image to _apply_ken_burns.
                    # But wait, _apply_ken_burns resizes it AGAIN.
                    # Let's optimize:
                    # Pass the original image (or slightly downscaled if huge) to _apply_ken_burns?
                    # No, let's stick to the pattern:
                    # 1. Resize to 'Cover' (this ensures we have enough pixels to fill screen)
                    # 2. Pass this to _apply_ken_burns.
                    # 3. _apply_ken_burns will zoom in FURTHER (1.2x etc).
                    
                    # However, if we resize to cover, we might crop out the face if the aspect ratio is very different.
                    # Example: Portrait image (9:16) on Landscape video (16:9).
                    # 'Cover' will crop top/bottom or left/right.
                    # For Portrait on Landscape: Width matches, Height is huge. We crop top/bottom.
                    # We must ensure the crop includes the face.
                    
                    # Let's do a smart crop to 'Cover' resolution first, centered on focus point.
                    
                    # Target dimensions
                    tw, th = target_w, target_h
                    
                    # Current dimensions
                    cw, ch = new_w, new_h
                    
                    # We need to crop (cw, ch) to (tw, th)
                    # Focus point in (cw, ch)
                    fx_px = focus_x * cw
                    fy_px = focus_y * ch
                    
                    # Top-left of crop
                    left = fx_px - tw / 2
                    top = fy_px - th / 2
                    
                    # Clamp
                    left = max(0, min(left, cw - tw))
                    top = max(0, min(top, ch - th))
                    
                    pil_img_smart_cropped = pil_img_resized.crop((left, top, left + tw, top + th))
                    
                    # Now we have an image exactly size of resolution, centered on face.
                    # Now apply Ken Burns (zoom/pan) on this.
                    # Note: Since we cropped tightly, we can only Zoom IN.
                    # If we want to Zoom OUT, we needed more context.
                    # But Zoom IN is safe.
                    
                    img_clip = ImageClip(np.array(pil_img_smart_cropped.convert("RGB"))).with_duration(duration_per_image)
                    
                    # Apply Effect (Zoom In/Out relative to center of THIS cropped image)
                    # Since we centered on face, the center of this image IS the face (roughly).
                    # So we can just zoom in/out of center.
                    kb_clip = self._apply_ken_burns(img_clip, target_w, target_h, focus_point=(0.5, 0.5))
                    
                    # Create Text Layer
                    txt_img = Image.new("RGBA", (target_w, target_h), (0, 0, 0, 0))
                    d = ImageDraw.Draw(txt_img)
                    
                    # Draw Title (only on first image)
                    if i == 0 and title_text:
                        print(f"Drawing title: {title_text}")
                        bbox = d.textbbox((0, 0), title_text, font=title_font)
                        text_w = bbox[2] - bbox[0]
                        text_h = bbox[3] - bbox[1]
                        x = (target_w - text_w) / 2
                        y = target_h - text_h - 100
                        d.text((x+2, y+2), title_text, font=title_font, fill=(0, 0, 0, 200))
                        d.text((x, y), title_text, font=title_font, fill=(255, 255, 255, 255))
                        
                    # Draw Date
                    bbox = d.textbbox((0, 0), date_str, font=date_font)
                    text_w = bbox[2] - bbox[0]
                    text_h = bbox[3] - bbox[1]
                    x = target_w - text_w - 30
                    y = target_h - text_h - 30
                    d.text((x+2, y+2), date_str, font=date_font, fill=(0, 0, 0, 200))
                    d.text((x, y), date_str, font=date_font, fill=(255, 255, 255, 255))
                    
                    txt_clip = ImageClip(np.array(txt_img)).with_duration(duration_per_image)
                    
                    # Composite
                    clip = CompositeVideoClip([kb_clip, txt_clip])
                    
                else:
                    # STATIC LOGIC (Original)
                    # Scale to fit (contain)
                    scale = min(target_w / w, target_h / h)
                    new_w = int(w * scale)
                    new_h = int(h * scale)
                    
                    pil_img = pil_img.resize((new_w, new_h), Image.Resampling.LANCZOS)
                    
                    # Create text layer
                    txt_img = Image.new("RGBA", pil_img.size, (255, 255, 255, 0))
                    d = ImageDraw.Draw(txt_img)
                    
                    # Update dimensions
                    w, h = pil_img.size
                    
                    # Draw Title (only on first image)
                    if i == 0:
                        print(f"Processing first image. Title text: '{title_text}'")
                        if title_text:
                            print(f"Drawing title: {title_text}")
                            bbox = d.textbbox((0, 0), title_text, font=title_font)
                            text_w = bbox[2] - bbox[0]
                            text_h = bbox[3] - bbox[1]
                            
                            x = (w - text_w) / 2
                            y = h - text_h - 100 # 100px padding from bottom
                            
                            # Draw shadow
                            d.text((x+2, y+2), title_text, font=title_font, fill=(0, 0, 0, 200))
                            # Draw text
                            d.text((x, y), title_text, font=title_font, fill=(255, 255, 255, 255))
                    
                    # Draw Date (on all images)
                    bbox = d.textbbox((0, 0), date_str, font=date_font)
                    text_w = bbox[2] - bbox[0]
                    text_h = bbox[3] - bbox[1]
                    
                    x = w - text_w - 30
                    y = h - text_h - 30
                    
                    # Draw shadow
                    d.text((x+2, y+2), date_str, font=date_font, fill=(0, 0, 0, 200))
                    # Draw text
                    d.text((x, y), date_str, font=date_font, fill=(255, 255, 255, 255))
                    
                    out = Image.alpha_composite(pil_img, txt_img)
                    clip = ImageClip(np.array(out.convert("RGB"))).with_duration(duration_per_image)
                    
                    # Center (MoviePy handles positioning if clip is smaller than screen, 
                    # but we usually want to fill or fit. Here we fit.)
                    clip = clip.with_position("center")
                
                # Apply fade in/out for smooth transitions
                clip = clip.with_effects([vfx.FadeIn(0.5), vfx.FadeOut(0.5)])
                
                clips.append(clip)
            
            # Concatenate clips
            # method="compose" is safer for different sizes
            final_clip = concatenate_videoclips(clips, method="compose")
            
            # Add audio
            if audio_path and os.path.exists(audio_path):
                audio = AudioFileClip(audio_path)
                
                # Handle Trimming
                if audio_start > 0 or (audio_end is not None and audio_end > 0):
                    start = audio_start
                    end = audio_end if audio_end is not None and audio_end > 0 else audio.duration
                    # Ensure end is within bounds
                    end = min(end, audio.duration)
                    if start < end:
                        audio = audio.subclipped(start, end)
                
                # Loop audio if shorter than video
                if audio.duration < final_clip.duration:
                    # Manual loop
                    n_loops = int(final_clip.duration / audio.duration) + 1
                    audio = concatenate_audioclips([audio] * n_loops).subclipped(0, final_clip.duration)
                else:
                    audio = audio.subclipped(0, final_clip.duration)
                final_clip = final_clip.with_audio(audio)

            # Write file
            logger = None
            if progress_callback:
                logger = CallbackLogger(progress_callback)
            
            final_clip.write_videofile(output_path, fps=24, codec='libx264', audio_codec='aac', logger=logger or 'bar')
            return True
        except Exception as e:
            print(f"Error generating video: {e}")
            import traceback
            traceback.print_exc()
            return False

    def _apply_ken_burns(self, clip, width, height, focus_point=(0.5, 0.5)):
        """
        Applies a Ken Burns effect (pan & zoom) to the clip.
        focus_point: (x, y) tuple in 0.0-1.0 range indicating interest point.
        """
        w, h = clip.size
        duration = clip.duration
        
        # Zoom factor
        zoom_factor = random.uniform(1.15, 1.35)
        
        # Calculate dimensions of the zoomed image
        new_w = int(width * zoom_factor)
        new_h = int(height * zoom_factor)
        
        # Resize the clip first
        zoomed_clip = clip.resized((new_w, new_h))
        
        # Available movement range
        max_x = new_w - width
        max_y = new_h - height
        
        # Focus point in zoomed clip pixels
        fx = focus_point[0] * new_w
        fy = focus_point[1] * new_h
        
        # Ideal top-left to center the focus point
        target_x = fx - width / 2
        target_y = fy - height / 2
        
        # Clamp to bounds
        target_x = max(0, min(target_x, max_x))
        target_y = max(0, min(target_y, max_y))
        
        # Determine movement
        # 70% Zoom In, 30% Zoom Out
        mode = 'zoom_in' if random.random() < 0.7 else 'zoom_out'
        
        if mode == 'zoom_in':
            # Start: Center of image (or slightly offset)
            # End: target_x, target_y (Focused on face)
            
            # Start at center of scrollable area
            start_x = max_x / 2
            start_y = max_y / 2
            
            # Or start at random position?
            # Let's start at center for stability
            
            end_x = target_x
            end_y = target_y
            
        else: # zoom_out
            # Start: Focused on face
            # End: Center
            
            start_x = target_x
            start_y = target_y
            
            end_x = max_x / 2
            end_y = max_y / 2

        def make_frame(t):
            # Ease-in-out interpolation? Or just linear.
            # Linear is standard for Ken Burns.
            progress = t / duration
            
            # Simple ease-out for smoother feel?
            # progress = 1 - (1 - progress) * (1 - progress)
            
            x = int(start_x + (end_x - start_x) * progress)
            y = int(start_y + (end_y - start_y) * progress)
            
            # Ensure bounds
            x = max(0, min(x, max_x))
            y = max(0, min(y, max_y))
            
            # Get frame from zoomed clip
            frame = zoomed_clip.get_frame(t)
            
            # Crop: [y:y+height, x:x+width]
            return frame[y:y+height, x:x+width]

        return VideoClip(make_frame, duration=duration)




from moviepy.editor import AudioFileClip

class AudioManager:
    def __init__(self):
        pass

    def get_audio_duration(self, audio_path: str):
        try:
            clip = AudioFileClip(audio_path)
            duration = clip.duration
            clip.close()
            return duration
        except Exception as e:
            print(f"Error reading audio {audio_path}: {e}")
            return 0

    def trim_audio(self, input_path: str, output_path: str, start_time: float, end_time: float):
        try:
            clip = AudioFileClip(input_path)
            trimmed = clip.subclip(start_time, end_time)
            trimmed.write_audiofile(output_path)
            clip.close()
            return True
        except Exception as e:
            print(f"Error trimming audio: {e}")
            return False



from setuptools import setup, find_packages

setup(
    name="video_highlight_generator",
    version="0.1.1",
    packages=find_packages(where="backend"),
    package_dir={"": "backend"},
    include_package_data=True,
    install_requires=[
        "fastapi",
        "uvicorn",
        "opencv-python",
        "torch",
        "facenet-pytorch",
        "moviepy>=2.0.0.dev0",
        "numpy>=2.0.0",
        "Pillow",
        "python-multipart",
        "aiofiles"
    ],
    entry_points={
        "console_scripts": [
            "video-highlight-generator=video_highlight_generator.main:start",
        ],
    },
)
